import sys
import time
import math
import serial
import threading
from PyQt5.QtWidgets import (QApplication, QWidget, QPushButton, QVBoxLayout, 
                             QHBoxLayout, QGroupBox, QLabel, QLineEdit, QProgressBar, 
                             QComboBox, QPlainTextEdit, QFrame, QSpacerItem, QSizePolicy)
from PyQt5.QtCore import (QTimer, Qt, pyqtSignal, QRectF)
from PyQt5.QtGui import QPainter, QPen, QColor, QFont
from serial.tools import list_ports

class CircularTimerWidget(QWidget):
    def __init__(self, total_seconds=5, arc_color=QColor(0, 122, 255)):
        super().__init__()
        self.total_seconds = total_seconds
        self.arc_color = arc_color

        self.setFixedSize(100, 100)
        self.setAttribute(Qt.WA_TranslucentBackground)

        self.timer = QTimer(self)
        self.timer.timeout.connect(self.update_timer)
        self.running = False

    def start_countdown(self, seconds=None):
        if seconds is not None:
            self.total_seconds = seconds
        self.start_time = time.time()
        self.end_time = self.start_time + self.total_seconds
        self.running = True
        self.timer.start(20)  # Update every 20ms for smoother arc movement

    def stop_countdown(self):
        self.timer.stop()
        self.running = False
        self.update()

    def update_timer(self):
        now = time.time()
        remaining = self.end_time - now
        if remaining <= 0:
            self.running = False
            self.timer.stop()
            self.update()
        else:
            self.update()

    def paintEvent(self, event):
        painter = QPainter(self)
        painter.setRenderHint(QPainter.Antialiasing, True)

        size = min(self.width(), self.height())
        margin = 10
        arc_rect = QRectF(margin, margin, size - 2*margin, size - 2*margin)

        # Draw background circle
        background_pen = QPen(Qt.lightGray, 8, Qt.SolidLine, Qt.RoundCap, Qt.RoundJoin)
        painter.setPen(background_pen)
        painter.setBrush(Qt.NoBrush)
        painter.drawEllipse(arc_rect)

        if self.running:
            now = time.time()
            remaining = self.end_time - now
            if remaining < 0:
                remaining = 0
            fraction = remaining / self.total_seconds
            angle = 360.0 * fraction

            # Draw arc
            arc_pen = QPen(self.arc_color, 8, Qt.SolidLine, Qt.RoundCap, Qt.RoundJoin)
            painter.setPen(arc_pen)
            start_angle = 90 * 16
            span_angle = -int(angle * 16)
            painter.drawArc(arc_rect, start_angle, span_angle)

            painter.setPen(Qt.white)
            painter.setFont(QFont("Arial", 14, QFont.Bold))
            sec_str = str(int(math.ceil(remaining)))
            painter.drawText(arc_rect, Qt.AlignCenter, sec_str)

class SerialReaderThread(threading.Thread):
    def __init__(self, ser, line_callback, debug_callback):
        super().__init__()
        self.ser = ser
        self.line_callback = line_callback
        self.debug_callback = debug_callback
        self._stop_flag = False
        self.last_data_time = time.time()

    def run(self):
        self.debug_callback("Reader thread started. Waiting for data...\n")
        while not self._stop_flag:
            if self.ser.in_waiting > 0:
                line = self.ser.readline().decode(errors='replace')
                if line:
                    line = line.strip()
                    print("Incoming line:", line)
                    self.line_callback(line)
                    self.last_data_time = time.time()
            else:
                if time.time() - self.last_data_time > 5:
                    self.debug_callback("No data received in the last 5 seconds.\n")
                    self.last_data_time = time.time()
            time.sleep(0.01)

    def stop(self):
        self._stop_flag = True

class MainWindow(QWidget):
    transducerLineReceived = pyqtSignal(str)
    piezoLineReceived = pyqtSignal(str)
    transducerDebug = pyqtSignal(str)
    piezoDebug = pyqtSignal(str)

    def __init__(self):
        super().__init__()
        self.setWindowTitle("Dual Arduino Control Panel")

        self.transducer_ser = None
        self.piezo_ser = None
        self.transducer_thread = None
        self.piezo_thread = None

        self.FSR_MAX = 900
        self.led_ranges = [150, 300, 450, 600, 750, 900]

        self.transducerLineReceived.connect(self.onTransducerLine)
        self.piezoLineReceived.connect(self.onPiezoLine)
        self.transducerDebug.connect(self.onTransducerDebug)
        self.piezoDebug.connect(self.onPiezoDebug)

        self.a_count = 0
        self.p_count = 0  # For the 'p' key sample counter

        # Timers
        self.transducerTimerWidget = CircularTimerWidget(total_seconds=3, arc_color=QColor(0,122,255))
        self.piezoTimerWidget = CircularTimerWidget(total_seconds=5, arc_color=QColor(0,200,0))

        # Automate timer
        self.automateTimer = QTimer()
        self.automateTimer.setSingleShot(True)
        self.automateTimer.timeout.connect(self.sendTransducerSweep)

        self.recordingStoppedCheckVisible = False

        self.initUI()

        self.uiTimer = QTimer()
        self.uiTimer.timeout.connect(self.updateUI)
        self.uiTimer.start(100)

        # Capture key events for 'c' and 'p'
        self.setFocusPolicy(Qt.StrongFocus)

    def initUI(self):
        layout = QVBoxLayout(self)

        # Connection Controls
        portLayout = QHBoxLayout()
        self.available_baudrates = ["9600", "19200", "38400", "57600", "115200"]
        default_baud = "115200"

        # Transducer port
        transducerSelectLayout = QVBoxLayout()
        transducerSelectLayout.addWidget(QLabel("Transducer Port:"))
        self.transducerCombo = QComboBox()
        self.transducerCombo.setEditable(True)
        transducerSelectLayout.addWidget(self.transducerCombo)
        transducerSelectLayout.addWidget(QLabel("Baud Rate:"))
        self.transducerBaudCombo = QComboBox()
        for br in self.available_baudrates:
            self.transducerBaudCombo.addItem(br)
        self.transducerBaudCombo.setCurrentText(default_baud)
        transducerSelectLayout.addWidget(self.transducerBaudCombo)
        self.connectTransducerBtn = QPushButton("Connect Transducer")
        self.connectTransducerBtn.clicked.connect(self.connectTransducer)
        transducerSelectLayout.addWidget(self.connectTransducerBtn)

        # Piezo port
        piezoSelectLayout = QVBoxLayout()
        piezoSelectLayout.addWidget(QLabel("Piezo Port:"))
        self.piezoCombo = QComboBox()
        self.piezoCombo.setEditable(True)
        piezoSelectLayout.addWidget(self.piezoCombo)
        piezoSelectLayout.addWidget(QLabel("Baud Rate:"))
        self.piezoBaudCombo = QComboBox()
        for br in self.available_baudrates:
            self.piezoBaudCombo.addItem(br)
        self.piezoBaudCombo.setCurrentText(default_baud)
        piezoSelectLayout.addWidget(self.piezoBaudCombo)
        self.connectPiezoBtn = QPushButton("Connect Piezo")
        self.connectPiezoBtn.clicked.connect(self.connectPiezo)
        piezoSelectLayout.addWidget(self.connectPiezoBtn)

        refreshBtn = QPushButton("Refresh Ports")
        refreshBtn.clicked.connect(self.refreshPorts)

        portLayout.addLayout(transducerSelectLayout)
        portLayout.addLayout(piezoSelectLayout)
        portLayout.addWidget(refreshBtn)
        layout.addLayout(portLayout)
        self.refreshPorts()

        # Transducer Side
        transducerGroup = QGroupBox("Transducer Side (FSR Gauge)")
        transducerLayout = QVBoxLayout()

        self.transducerTerminal = QPlainTextEdit()
        self.transducerTerminal.setReadOnly(True)
        transducerLayout.addWidget(self.transducerTerminal)

        # Transducer command line
        transducerCmdLayout = QHBoxLayout()
        self.transducerCmdInput = QLineEdit()
        transducerSendBtn = QPushButton("Send")
        transducerSendBtn.clicked.connect(self.sendTransducerCommand)
        transducerCmdLayout.addWidget(self.transducerCmdInput)
        transducerCmdLayout.addWidget(transducerSendBtn)
        transducerLayout.addLayout(transducerCmdLayout)

        # FSR and LED boxes
        fsrDisplayLayout = QHBoxLayout()
        fsrGaugeLayout = QVBoxLayout()

        # Put reading and percentage on the same line
        self.fsrLabel = QLabel("FSR Reading: 0 (0%)")
        fsrGaugeLayout.addWidget(self.fsrLabel)

        # Style the FSR gauge: red with black outline, rounded
        self.fsrGauge = QProgressBar()
        self.fsrGauge.setMinimum(0)
        self.fsrGauge.setMaximum(self.FSR_MAX)
        self.fsrGauge.setValue(0)
        self.fsrGauge.setStyleSheet("""
            QProgressBar {
                border: 2px solid black;
                border-radius: 5px;
                background: none;
            }
            QProgressBar::chunk {
                background: red;
                border-radius: 5px;
            }
        """)
        fsrGaugeLayout.addWidget(QLabel("FSR Gauge:"))
        fsrGaugeLayout.addWidget(self.fsrGauge)

        self.ledCountLabel = QLabel("LEDs Lit: 0")
        fsrGaugeLayout.addWidget(self.ledCountLabel)

        # LED boxes - change to dark green when lit (instead of light blue), gray otherwise, red if ≥6, rounded
        ledBoxLayout = QHBoxLayout()
        self.ledBoxes = []
        for i in range(5):
            box = QFrame()
            box.setFixedSize(20,20)
            box.setStyleSheet("background-color: gray; border:1px solid black; border-radius:5px;")
            self.ledBoxes.append(box)
            ledBoxLayout.addWidget(box)
        fsrGaugeLayout.addLayout(ledBoxLayout)

        fsrDisplayLayout.addLayout(fsrGaugeLayout)
        transducerLayout.addLayout(fsrDisplayLayout)
        transducerGroup.setLayout(transducerLayout)

        # Piezo Side
        piezoGroup = QGroupBox("Piezo Side")
        piezoLayout = QVBoxLayout()

        # Add a horizontal layout for terminal and a checkmark label
        piezoTopLayout = QHBoxLayout()
        self.piezoTerminal = QPlainTextEdit()
        self.piezoTerminal.setReadOnly(True)
        piezoTopLayout.addWidget(self.piezoTerminal)

        # Checkmark label (hidden by default)
        self.piezoCheckLabel = QLabel("")
        self.piezoCheckLabel.setStyleSheet("color: gold; font-size: 20px;")
        piezoTopLayout.addWidget(self.piezoCheckLabel)

        piezoLayout.addLayout(piezoTopLayout)

        # Piezo command line
        piezoCmdLayout = QHBoxLayout()
        self.piezoCmdInput = QLineEdit()
        piezoSendBtn = QPushButton("Send")
        piezoSendBtn.clicked.connect(self.sendPiezoCommand)
        piezoCmdLayout.addWidget(self.piezoCmdInput)
        piezoCmdLayout.addWidget(piezoSendBtn)
        piezoLayout.addLayout(piezoCmdLayout)

        piezoGroup.setLayout(piezoLayout)

        deviceLayout = QHBoxLayout()
        deviceLayout.addWidget(transducerGroup)
        deviceLayout.addWidget(piezoGroup)
        layout.addLayout(deviceLayout)

        # Timers above Quick Commands
        timersLayout = QHBoxLayout()
        # Transducer timer
        transTimerLayout = QVBoxLayout()
        transTimerLayout.addWidget(self.transducerTimerWidget, 0, Qt.AlignCenter)
        tLabel = QLabel("Transducer Timer")
        tLabel.setStyleSheet("color: white;")
        tLabel.setAlignment(Qt.AlignCenter)
        transTimerLayout.addWidget(tLabel, 0, Qt.AlignCenter)
        transTimerContainer = QWidget()
        transTimerContainer.setLayout(transTimerLayout)

        # Piezo timer
        piezoTimerLayout = QVBoxLayout()
        piezoTimerLayout.addWidget(self.piezoTimerWidget, 0, Qt.AlignCenter)
        pLabel = QLabel("Piezo Timer")
        pLabel.setStyleSheet("color: white;")
        pLabel.setAlignment(Qt.AlignCenter)
        piezoTimerLayout.addWidget(pLabel, 0, Qt.AlignCenter)
        piezoTimerContainer = QWidget()
        piezoTimerContainer.setLayout(piezoTimerLayout)

        timersLayout.addWidget(transTimerContainer)
        timersLayout.addSpacerItem(QSpacerItem(50,50,QSizePolicy.Fixed,QSizePolicy.Fixed))
        timersLayout.addWidget(piezoTimerContainer)

        layout.addLayout(timersLayout)

        # Quick Commands
        hotkeyGroup = QGroupBox("Quick Commands")
        hotkeyLayout = QHBoxLayout()
        self.calibrateLowBtn = QPushButton("Calibrate Low (y)")
        self.calibrateLowBtn.clicked.connect(self.handleCalibrateLow)
        hotkeyLayout.addWidget(self.calibrateLowBtn)
        self.calibrateHighBtn = QPushButton("Calibrate High (y after low)")
        self.calibrateHighBtn.clicked.connect(self.handleCalibrateHigh)
        hotkeyLayout.addWidget(self.calibrateHighBtn)
        self.resetLowBtn = QPushButton("Reset Low (r)")
        self.resetLowBtn.clicked.connect(lambda: self.sendTransducerCommandString("r"))
        hotkeyLayout.addWidget(self.resetLowBtn)
        self.startSweepBtn = QPushButton("Start Sweep (s)")
        self.startSweepBtn.clicked.connect(self.handleStartSweep)
        hotkeyLayout.addWidget(self.startSweepBtn)

        self.automateBtn = QPushButton("Automate (a)")
        self.automateBtn.clicked.connect(self.handleAutomateAction)
        hotkeyLayout.addWidget(self.automateBtn)
        hotkeyGroup.setLayout(hotkeyLayout)
        layout.addWidget(hotkeyGroup)

        # Automate and Sample counts
        self.aCountLabel = QLabel("Automate actions: 0")
        self.samplesLabel = QLabel("Samples: 0")
        layout.addWidget(self.aCountLabel)
        layout.addWidget(self.samplesLabel)

        self.setLayout(layout)
        self.resize(1200, 700)

    def refreshPorts(self):
        self.transducerCombo.clear()
        self.piezoCombo.clear()
        ports = list_ports.comports()
        for p in ports:
            self.transducerCombo.addItem(p.device)
            self.piezoCombo.addItem(p.device)

    def keyPressEvent(self, event):
        if event.key() == Qt.Key_C:
            self.handleAutomateAction()
        elif event.key() == Qt.Key_P:
            self.p_count += 1
            self.samplesLabel.setText(f"Samples: {self.p_count}")
        super().keyPressEvent(event)

    def connectTransducer(self):
        port = self.transducerCombo.currentText().strip()
        baud = int(self.transducerBaudCombo.currentText().strip())
        if not port:
            self.appendTransducer("No port selected for Transducer.")
            return
        try:
            self.transducer_ser = serial.Serial(port, baud, timeout=0.1, dsrdtr=False, rtscts=False)
            time.sleep(2)
            self.appendTransducer(f"Connected to Transducer at {port}")
            self.transducer_thread = SerialReaderThread(
                self.transducer_ser,
                line_callback=self.transducerLineReceived.emit,
                debug_callback=self.transducerDebug.emit
            )
            self.transducer_thread.start()
        except Exception as e:
            self.appendTransducer("Failed to connect: " + str(e))

    def connectPiezo(self):
        port = self.piezoCombo.currentText().strip()
        baud = int(self.piezoBaudCombo.currentText().strip())
        if not port:
            self.appendPiezo("No port selected for Piezo.")
            return
        try:
            self.piezo_ser = serial.Serial(port, baud, timeout=0.1, dsrdtr=False, rtscts=False)
            time.sleep(2)
            self.appendPiezo("Connected to Piezo at " + port)
            self.piezo_thread = SerialReaderThread(
                self.piezo_ser,
                line_callback=self.piezoLineReceived.emit,
                debug_callback=self.piezoDebug.emit
            )
            self.piezo_thread.start()
        except Exception as e:
            self.appendPiezo("Failed to connect: " + str(e))

    def onTransducerDebug(self, msg):
        self.appendTransducer("DEBUG: " + msg.strip())

    def onPiezoDebug(self, msg):
        self.appendPiezo("DEBUG: " + msg.strip())

    def onTransducerLine(self, line):
        self.appendTransducer(line)
        try:
            val = int(line)
            if 0 <= val <= 1023:
                self.updateFSR(val)
        except ValueError:
            pass

    def onPiezoLine(self, line):
        self.appendPiezo(line)
        if "Recording started" in line:
            self.piezoTimerWidget.start_countdown(5)
        if "Recording stopped" in line:
            # Show a gold checkmark
            self.piezoCheckLabel.setText("✔")
            self.recordingStoppedCheckVisible = True

    def updateFSR(self, val):
        fsr_val = min(val, self.FSR_MAX)
        percentage = (fsr_val / self.FSR_MAX) * 100
        self.fsrLabel.setText(f"FSR Reading: {fsr_val} ({int(percentage)}%)")
        self.fsrGauge.setValue(fsr_val)
        self.updateLEDCount(fsr_val)

    def appendTransducer(self, text):
        self.transducerTerminal.appendPlainText(text)

    def appendPiezo(self, text):
        self.piezoTerminal.appendPlainText(text)

    def sendTransducerCommand(self):
        cmd = self.transducerCmdInput.text().strip()
        if cmd and self.transducer_ser and self.transducer_ser.is_open:
            self.transducer_ser.write((cmd + "\n").encode())
            self.appendTransducer(">> " + cmd)
            self.transducerCmdInput.clear()

    def sendPiezoCommand(self):
        cmd = self.piezoCmdInput.text().strip()
        if cmd and self.piezo_ser and self.piezo_ser.is_open:
            self.piezo_ser.write((cmd + "\n").encode())
            self.appendPiezo(">> " + cmd)
            self.piezoCmdInput.clear()
            if cmd.lower() == 'a':
                self.triggerAutomateAction()

    def sendTransducerCommandString(self, cmd):
        if self.transducer_ser and self.transducer_ser.is_open:
            self.transducer_ser.write((cmd + "\n").encode())
            self.appendTransducer(">> " + cmd)

    def handleCalibrateLow(self):
        self.sendTransducerCommandString("y")
        self.startTransducerTimer(5)

    def handleCalibrateHigh(self):
        self.sendTransducerCommandString("y")
        self.startTransducerTimer(5)

    def handleStartSweep(self):
        self.sendTransducerCommandString("s")
        self.startTransducerTimer(3)

    def handleAutomateAction(self):
        if self.piezo_ser and self.piezo_ser.is_open:
            self.piezo_ser.write("a\n".encode())
            self.appendPiezo(">> a")
            self.triggerAutomateAction()
        else:
            self.appendPiezo("Piezo not connected, cannot send 'a'.")

    def triggerAutomateAction(self):
        # Hide the checkmark if it was visible
        if self.recordingStoppedCheckVisible:
            self.piezoCheckLabel.setText("")
            self.recordingStoppedCheckVisible = False

        # Start both timers on automate with a 0.5s delay for transducer
        QTimer.singleShot(500, lambda: self.startTransducerTimer(3))
        self.piezoTimerWidget.start_countdown(5)

        self.a_count += 1
        self.aCountLabel.setText(f"Automate actions: {self.a_count}")
        self.automateTimer.start(500)

    def sendTransducerSweep(self):
        self.sendTransducerCommandString("s")

    def startTransducerTimer(self, seconds):
        self.transducerTimerWidget.start_countdown(seconds)

    def updateLEDCount(self, val):
        led_count = 0
        for threshold in self.led_ranges:
            if val > threshold:
                led_count += 1
            else:
                break
        if val > 0 and led_count == 0:
            led_count = 1

        # If led_count ≥ 6 → all red
        # If led_count ≤ 5 → that many green (#228B22) from left, rest gray
        # If led_count = 0 → all gray
        if led_count >= 6:
            # All red
            for box in self.ledBoxes:
                box.setStyleSheet("background-color: red; border:1px solid black; border-radius:5px;")
        else:
            for i in range(5):
                if i < led_count:
                    box_color = "#228B22"  # dark green
                else:
                    box_color = "gray"
                self.ledBoxes[i].setStyleSheet(f"background-color: {box_color}; border:1px solid black; border-radius:5px;")

        self.ledCountLabel.setText(f"LEDs Lit: {led_count}")

    def updateUI(self):
        pass

    def closeEvent(self, event):
        if self.transducer_thread:
            self.transducer_thread.stop()
        if self.piezo_thread:
            self.piezo_thread.stop()
        if self.transducer_ser and self.transducer_ser.is_open:
            self.transducer_ser.close()
        if self.piezo_ser and self.piezo_ser.is_open:
            self.piezo_ser.close()
        super().closeEvent(event)

if __name__ == '__main__':
    app = QApplication(sys.argv)
    w = MainWindow()
    w.show()
    sys.exit(app.exec_())
